#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include "prngd.h"

static volatile seed_t gatherer_seed;
static volatile pid_t sigchild_pid = -1;
static volatile int gatherer_status;
static volatile int caught_sigchild = 0;
static volatile int received_sighup = 0;

typedef enum
{
  wait_command, wait_second_byte, wait_third_byte, wait_fourth_byte,
  wait_read, wait_write, wait_entropy, to_be_closed, to_be_error_closed
}
querystate_t;

typedef struct
{
  int fd;
  int state;
  int to_process;
  int yet_processed;
  int entropy_bits;
  unsigned char command, secbyte;
  unsigned char buffer[256];
} egd_query_t;


static void termination_handler(int received_signal /*ARGSUSED*/)
{
  if (debug)
    (void)fprintf(stderr, "Caught SIGTERM, shutting down\n");
  close_seedfile();
  exit(EXIT_SUCCESS);
}

static void sighup_handler(int received_signal /*ARGSUSED*/)
{
  received_sighup = 1;
}

static void child_handler(int received_signal /*ARGSUSED*/)
{
  int save_errno = errno;
  int my_gatherer_status;
  pid_t pid;

  /*
   * Wait for child process to avoid zombies. This signal comes in
   * asynchronously and only few functions are safe to be called,
   * especially stdio and syslog functions cannot be called.
   *
   * PRNGD can only have one child at a time, so recording the information
   * is trivial.
   */
  caught_sigchild = 1;

  do {
    pid = waitpid(-1, &my_gatherer_status, WNOHANG);
  } while (pid == -1 && errno == EINTR);

  if (WIFEXITED(my_gatherer_status) || WIFSIGNALED(my_gatherer_status)) {
    sigchild_pid = pid;
    gatherer_status = my_gatherer_status;
  }

  errno = save_errno;
}

void main_loop(int *service_socket, int numsock, time_t seed_stat_interval,
	       time_t seed_ext_interval, entropy_source_t *entropy_source,
	       int max_gatherer_bytes)
{
  int i;
  int accept_fd;
  struct sockaddr addr;
  SOCKLEN_T addrlen;
  int num, numbytes, error;
  int pipedes[2];
  time_t last_seed_stat = 0, last_seed_ext = 0;
  int max_query = 0;
  int alloc_query = 0;
  egd_query_t *egd_query = NULL;
  egd_query_t *egd_query_old;
  int max_fd;
  pid_t gatherer_pid, child_pid;
  int gather_fd;
  int gatherer_entry;
  int gatherer_bytes;
  int first_gather = 1;
  double est = 0.0;
  time_t gather_start;
  unsigned char gatherbuf[10240];
  char msg_buf[1024];
  struct sigaction new_action;
  fd_set rfds, wfds, efds;
  struct timeval timeout;
  int saved_errno;

  gatherer_pid = -1;
  gather_fd = -1;
  gatherer_entry = 0;

  gatherer_bytes = 0;		/* Initialize to make gcc happy */
  gather_start = time(NULL);	/* Initialize to make gcc happy */

  /*
   * Set up the signal handling. TERM signals shall lead to clean
   * shutdown of prngd (save the seed and exit).
   * HUP signals shall lead to a re-read of the config file.
   * SIGPIPE might occur when a peer shuts down the connection. Of course,
   * we don't want prngd to be affected in any way (it should simply close
   * the connection), so SIGPIPE shall be ignored.
   */
  (void)memset(&new_action, 0, sizeof(struct sigaction));
  new_action.sa_handler = termination_handler;
  (void)sigemptyset(&new_action.sa_mask);
  new_action.sa_flags |= SA_RESTART;
  (void)sigaction(SIGTERM, &new_action, NULL);

  (void)memset(&new_action, 0, sizeof(struct sigaction));
  new_action.sa_handler = sighup_handler;
  (void)sigemptyset(&new_action.sa_mask);
  new_action.sa_flags |= SA_RESTART;
  (void)sigaction(SIGHUP, &new_action, NULL);

  (void)memset(&new_action, 0, sizeof(struct sigaction));
  new_action.sa_handler = child_handler;
  (void)sigemptyset(&new_action.sa_mask);
  new_action.sa_flags |= SA_RESTART;
  (void)sigaction(SIGCHLD, &new_action, NULL);

  (void)memset(&new_action, 0, sizeof(struct sigaction));
  new_action.sa_handler = SIG_IGN;
  (void)sigemptyset(&new_action.sa_mask);
  new_action.sa_flags |= SA_RESTART;
  (void)sigaction(SIGPIPE, &new_action, NULL);

  for (;;)
  {
    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    FD_ZERO(&efds);
    max_fd = 0;
    /*
     * Do not arm select for the listener socket if we are not ready
     * to accept another connection.
     */
    if (count_openfd < limit_openfd - OPEN_FD_RESERVE) {
      for (i = 0; i < numsock; i++)
      {
        FD_SET(service_socket[i], &rfds);
        if (service_socket[i] > max_fd)
	max_fd = service_socket[i];
      }
    }
    if (gather_fd != -1)
    {
      FD_SET(gather_fd, &rfds);
#ifndef SCO3
      FD_SET(gather_fd, &efds);
#endif
      if (gather_fd > max_fd)
	max_fd = gather_fd;
    }
    for (i = 0; i < max_query; i++)
    {
      if (egd_query[i].fd > max_fd)
	max_fd = egd_query[i].fd;
      FD_SET(egd_query[i].fd, &efds);
      if ((egd_query[i].state == wait_command) ||
	  (egd_query[i].state == wait_second_byte) ||
	  (egd_query[i].state == wait_third_byte) ||
	  (egd_query[i].state == wait_fourth_byte) ||
	  (egd_query[i].state == wait_read))
	FD_SET(egd_query[i].fd, &rfds);
      else
	FD_SET(egd_query[i].fd, &wfds);

    }
    /*
     * If we just entered the loop and gather for the first time, we
     * don't want to wait.
     */
    if (first_gather || (rand_pool() < (ENTROPY_THRESHOLD))) {
      timeout.tv_sec = 0;
      timeout.tv_usec = 10;
    }
    else {
      timeout.tv_sec = (seed_stat_interval < seed_ext_interval ?
			 seed_stat_interval : seed_ext_interval);
      timeout.tv_usec = 0;
    }
    num = select(max_fd + 1, &rfds, &wfds, &efds, &timeout);
    saved_errno = errno;

    /*
     * Depending on the scheduling the SIGCHILD may already be old. This
     * does not matter however, since we only have one child at a time
     * and do not start another one before the old one has been killed
     * (or at least tried to with SIGKILL).
     */
    if (caught_sigchild) {
      caught_sigchild = 0;
      child_pid = sigchild_pid;		/* shut up warnings, as sigchild_pid */
					/* is volatile */
      if (debug)
	(void)fprintf(stderr,
		      "Caught SIGCHLD (pid = %ld, gatherer_pid = %ld)\n",
		      (long)child_pid, (long)gatherer_pid);
      if (child_pid == gatherer_pid)
	gatherer_pid = -1;
    }

    if (received_sighup) {
      received_sighup = 0;
      if (debug)
	(void)fprintf(stderr, "Received SIGHUP\n");
    }

    if (time(0) >= last_seed_stat + seed_stat_interval) {
      seed_stat();
      last_seed_stat = time(0);
    }

    seed_internal(NULL);

    if (num < 0) {
      if (saved_errno == EINTR)
        continue;		/* Ignore rfds and friends, since invalid */
      else {
	error_fatal("select() failed", saved_errno);
	exit(EXIT_FAILURE);
      }
    }

    if (gather_fd != -1) {
      if (FD_ISSET(gather_fd, &rfds)) {
	/*
	 * Bytes are coming back from the gatherer process. Read chunks of
	 * up to 10240 bytes and feed them into the PRNG
	 */
	numbytes = socket_read(gather_fd, gatherbuf, 10240);
	rand_add(gatherbuf, numbytes, est * numbytes);
	if (debug)
	  (void)fprintf(stderr, "Read %d bytes from gatherer\n", numbytes);
	gatherer_bytes += numbytes;
	if ((numbytes <= 0) || (gatherer_bytes >= max_gatherer_bytes)) {
	  if (debug)
	    (void)fprintf(stderr, "Closing connection to gatherer\n");
	  (void)close(gather_fd);
	  rand_add((int *)&gatherer_status, sizeof(gatherer_status), 0.0);
	  rand_add((seed_t *)&gatherer_seed, sizeof(seed_t), 0.25);
	  gather_fd = -1;
	  count_openfd--;
	}
      }
      else if (FD_ISSET(gather_fd, &efds)) {
	if (debug)
	  (void)fprintf(stderr, "Closing connection to gatherer (efds)\n");
	(void)close(gather_fd);
	gather_fd = -1;
	count_openfd--;
      }
      if (gather_fd == -1) {
	/*
	 * The formerly running gatherer has just been shut down for normal
	 * or error conditions. Choose the next one.
	 */
	gatherer_entry++;
        if (entropy_source[gatherer_entry].path == NULL) {
	  first_gather = 0;	/* done */
	  gatherer_entry = 0;
	}
      }
    }

    /*
     * If a gatherer has not finished in time, kill it!! Re-check, whether
     * the signal came in in between. This might happen, since we close
     * the connection when too many bytes where sent. If SIGCHILD is not
     * there and the time is over, we kill anyway. If the time is not over,
     * the check will be done at the next possible occation.
     */
    if (caught_sigchild) {
      caught_sigchild = 0;
      child_pid = sigchild_pid;		/* shut up warnings, as sigchild_pid */
					/* is volatile */
      if (debug)
	(void)fprintf(stderr,
		      "Caught SIGCHLD (pid = %ld, gatherer_pid = %ld)\n",
		      (long)child_pid, (long)gatherer_pid);
      if (child_pid == gatherer_pid)
	gatherer_pid = -1;
    }
    if ((gatherer_pid > 0) &&
	(difftime(time(NULL), gather_start) > seed_ext_interval / 2 + 1)) {
      if (debug)
	(void)fprintf(stderr, "Gathering time exceeded, kill process %ld\n",
		      (long)gatherer_pid);
      if (kill(gatherer_pid, SIGKILL) == -1) {
      	error_error("kill() failed", errno);
      }
      gatherer_pid = -1;
      if (gather_fd != -1) {
	(void)close(gather_fd);
	gather_fd = -1;
	count_openfd--;
	gatherer_entry++;
	if (entropy_source[gatherer_entry].path == NULL) {
	  first_gather = 0;	/* done */
	  gatherer_entry = 0;
	}
      }
    }

    /*
     * Call a gatherer, if it is time to do so. When starting up, call
     * all gatherers once to obtain initial seeding.
     */
    if (((time(0) >= last_seed_ext + seed_ext_interval) || first_gather ||
	(rand_pool() < (ENTROPY_THRESHOLD))) &&
	(gatherer_pid == -1) && (gather_fd == -1))
    {
      last_seed_ext = time(0);
      /*
       * time to call another gatherer 
       */
      if (my_pipe(pipedes) < 0)
	continue;		/*
				 * We cannot do anything about it 
				 */

      /*
       * Select the next gatherer to be used
       */
      if ((gatherer_entry == 0) &&
		(entropy_source[gatherer_entry].path == NULL)) {
        first_gather = 0;	/* no gatherer available */
        continue;
      }

      est = entropy_source[gatherer_entry].rate; /* for later use */
      gatherer_bytes = 0;

      gatherer_pid = fork();
      if (gatherer_pid == -1) {
	error_error("Could not fork", errno);
	(void)close(pipedes[0]);
	(void)close(pipedes[1]);
	continue;
      }
      if (gatherer_pid == 0) {
	/*
         * Child
	 */
	my_dup2(devnull, STDIN_FILENO);
        my_dup2(pipedes[1], STDOUT_FILENO);
        my_dup2(pipedes[1], STDERR_FILENO);
        (void)close(pipedes[0]);
        (void)close(pipedes[1]);
        (void)close(devnull); 
	if (execl(entropy_source[gatherer_entry].path,
	      entropy_source[gatherer_entry].args[0],
	      entropy_source[gatherer_entry].args[1],
	      entropy_source[gatherer_entry].args[2],
	      entropy_source[gatherer_entry].args[3],
	      entropy_source[gatherer_entry].args[4],
	      NULL) == -1)
	{
		(void)snprintf(msg_buf, sizeof(msg_buf),
			"Failed to execl(%s, %s, %s, %s, %s, %s)",
			entropy_source[gatherer_entry].path,
			entropy_source[gatherer_entry].args[0],
			entropy_source[gatherer_entry].args[1],
			entropy_source[gatherer_entry].args[2],
			entropy_source[gatherer_entry].args[3],
			entropy_source[gatherer_entry].args[4]);
		msg_buf[sizeof(msg_buf) - 1] = '\0';
		error_fatal(msg_buf, errno);
		_exit(EXIT_FAILURE);
	}
      }
      else {
	/*
	 * Parent
	 */
	(void)close(pipedes[1]);
	gather_fd = pipedes[0];
	count_openfd++;
	non_blocking(gather_fd, 1);
	gather_start = time(NULL);
	if (debug)
	{
	  (void)fprintf(stderr, "Spawned gatherer (pid = %ld): %s (%s",
		(long)gatherer_pid,
		entropy_source[gatherer_entry].path,
		entropy_source[gatherer_entry].args[0]);
	  i = 1;
	  while ((i < 5) && (entropy_source[gatherer_entry].args[i])) {
	    (void)fprintf(stderr, " %s",
	    	entropy_source[gatherer_entry].args[i++]);
	  }
	  (void)fprintf(stderr, ")\n");
	}
      }
      continue;

    }
    /*
     * Do service the active queries first, because accepting a new query
     * will create a new fd. First step: check errors reported and mark
     * the corresponding descriptor "to_be_error_closed".
     */
    for (i = 0; i < max_query; i++)
    {
      if (FD_ISSET(egd_query[i].fd, &efds))
      {
	if (debug)
	  (void)fprintf(stderr, "Error flagged for service fd %d\n",
	  	egd_query[i].fd);
	egd_query[i].state = to_be_error_closed;
      }
      else if (FD_ISSET(egd_query[i].fd, &rfds) ||
	       FD_ISSET(egd_query[i].fd, &wfds) ||
	       (egd_query[i].state == wait_entropy)) {
      switch (egd_query[i].state)
      {
      case wait_entropy:
	  seed_internal(NULL);	/* Mix in just before query */
	  seed_stat();
	  if (rand_bytes(egd_query[i].buffer, egd_query[i].secbyte) > 0)
	    egd_query[i].state = wait_write;
	break;
      case wait_command:
	error = socket_read(egd_query[i].fd, &(egd_query[i].command), 1);
	if (error == -1)
	  egd_query[i].state = to_be_error_closed;
	else if (error != 1)
	  egd_query[i].state = to_be_closed;
	else if (egd_query[i].command > 4)
	  egd_query[i].state = to_be_error_closed;
	else if (egd_query[i].command == 0)
	{
	  numbytes = rand_pool();
	  egd_query[i].buffer[0] = numbytes / 16777216;
	  egd_query[i].buffer[1] = (numbytes % 16777216) / 65536;
	  egd_query[i].buffer[2] = (numbytes % 65536) / 256;
	  egd_query[i].buffer[3] = numbytes % 256;
	  egd_query[i].to_process = 4;
	  egd_query[i].yet_processed = 0;
	  egd_query[i].state = wait_write;
	}
	else if ((egd_query[i].command == 1) ||
		 (egd_query[i].command == 2) ||
		 (egd_query[i].command == 3))
	  egd_query[i].state = wait_second_byte;
	else if (egd_query[i].command == 4)
	{
	  egd_query[i].buffer[0] = 12;
	  (void)sprintf((char *) (egd_query[i].buffer + 1), "%12ld", (long)getpid());
	  egd_query[i].to_process = 13;
	  egd_query[i].yet_processed = 0;
	  egd_query[i].state = wait_write;
	}
	break;
      case wait_second_byte:
	error = socket_read(egd_query[i].fd, &(egd_query[i].secbyte), 1);
	if (error != 1)
	  egd_query[i].state = to_be_error_closed;
	else if (egd_query[i].command == 1)
	{
	  seed_internal(NULL);	/* Mix in just before query */
	  seed_stat();
	  if (rand_bytes(egd_query[i].buffer + 1, egd_query[i].secbyte))
	    numbytes = egd_query[i].secbyte;
	  else
	    numbytes = 0;
	  egd_query[i].buffer[0] = (char)numbytes;
	  egd_query[i].to_process = numbytes + 1;
	  egd_query[i].yet_processed = 0;
	  egd_query[i].state = wait_write;
	}
	else if (egd_query[i].command == 2)
	{
	  egd_query[i].to_process = egd_query[i].secbyte;
	  egd_query[i].yet_processed = 0;
	  seed_internal(NULL);	/* Mix in just before query */
	  seed_stat();
	  if (rand_bytes(egd_query[i].buffer, egd_query[i].secbyte) > 0)
	    egd_query[i].state = wait_write;
	  else
	    egd_query[i].state = wait_entropy;
	}
	else if (egd_query[i].command == 3)
	{
	  egd_query[i].entropy_bits = egd_query[i].secbyte * 256;
	  egd_query[i].state = wait_third_byte;
	}
	break;
      case wait_third_byte:
	error = socket_read(egd_query[i].fd, &(egd_query[i].secbyte), 1);
	if (error != 1)
	  egd_query[i].state = to_be_error_closed;
	else
	{
	  egd_query[i].entropy_bits += egd_query[i].secbyte;
	  egd_query[i].state = wait_fourth_byte;
	}
	break;
      case wait_fourth_byte:
	error = socket_read(egd_query[i].fd, &(egd_query[i].secbyte), 1);
	if (error != 1)
	  egd_query[i].state = to_be_error_closed;
	else
	{
	  egd_query[i].to_process = egd_query[i].secbyte;
	  egd_query[i].yet_processed = 0;
	  egd_query[i].state = wait_read;
	}
	break;
      case wait_read:
	num = egd_query[i].to_process - egd_query[i].yet_processed;
	error = socket_read(egd_query[i].fd, egd_query[i].buffer, num);
	if (error <= 0)
	  egd_query[i].state = to_be_error_closed;
	else
	{
	  rand_add(egd_query[i].buffer, error,
		   (double) error * egd_query[i].entropy_bits / BITS_PER_BYTE /
		   egd_query[i].to_process);
	  egd_query[i].yet_processed += error;
	  if (egd_query[i].yet_processed >= egd_query[i].to_process)
	    egd_query[i].state = wait_command;
	}
	break;
      case wait_write:
	error = socket_write(egd_query[i].fd,
		      &(egd_query[i].buffer[egd_query[i].yet_processed]),
		      egd_query[i].to_process -
		      egd_query[i].yet_processed);
	if (error <= 0)
	  egd_query[i].state = to_be_error_closed;
	else
	  egd_query[i].yet_processed += error;
	if (egd_query[i].yet_processed == egd_query[i].to_process) {
	  egd_query[i].state = wait_command;
	  (void)fprintf(stderr, "Sent %d bytes to service %d\n",
		egd_query[i].to_process, egd_query[i].fd);
		}
	break;
      default:
	break;
      }
      }
    }
    for (i = 0; i < max_query; i++) {
      if ((egd_query[i].state == to_be_closed) ||
	  (egd_query[i].state == to_be_error_closed)) {
	if (debug)
	  (void)fprintf(stderr, "Closing service fd %d, error=%d\n",
			  egd_query[i].fd,
			  (egd_query[i].state == to_be_error_closed));
	else if (egd_query[i].state == to_be_error_closed) {
	  (void)snprintf(msg_buf, sizeof(msg_buf),
	  	"closing service fd %d for error.", egd_query[i].fd);
	  msg_buf[sizeof(msg_buf) - 1] = '\0';
	  error_notice(msg_buf, 0);
	}
	(void)close(egd_query[i].fd);
	count_openfd--;
	/* Remove delivered entropy */
	(void)memset(egd_query[i].buffer, 0, 256);
	if (i < max_query - 1)
	{
	  (void)memmove(egd_query + i, egd_query + max_query - 1,
		  sizeof(egd_query_t));
	  i--;
	}
	max_query--;
      }
    }

    for (i = 0; i < numsock; i++)
    {
      if (count_openfd >= limit_openfd - OPEN_FD_RESERVE) {
	(void)snprintf(msg_buf, sizeof(msg_buf),
		"could not accept connection: no more slots (%d of %d)\n",
		count_openfd, limit_openfd - OPEN_FD_RESERVE);
	msg_buf[sizeof(msg_buf) - 1] = '\0';
	error_notice(msg_buf, 0);
	continue;	/* do not accept new connection */
      }
      if (FD_ISSET(service_socket[i], &rfds))
      {
        addrlen = sizeof(addr);
        accept_fd = accept(service_socket[i], &addr, &addrlen);
        if (accept_fd < 0)
	  continue;		/*
				 * Ignore and stay alive 
				 */

	count_openfd++;
	non_blocking(accept_fd, 1);
	close_on_exec(accept_fd, 1);

        /*
         * Create a new structure for the incoming request 
         */
        if (max_query + 1 > alloc_query)
        {
	  egd_query_old = egd_query;
	  if (alloc_query == 0)
	    egd_query = malloc((max_query + 1) * sizeof(egd_query_t));
	  else
	    egd_query =
	        realloc(egd_query_old,
		        (max_query + 1) * sizeof(egd_query_t));
	  if (!egd_query)
	  {
	    error_error("realloc() failed", errno);
	    (void)close(accept_fd);	/*
				 * Better this than complete shut down 
				 */
	    egd_query = egd_query_old;
	    continue;
	  }
	  alloc_query = max_query + 1;
        }
        /*
         * This routine is safe here, as in case of error we do "continue;"
         * in case of problems. So we have succesfully reached this point.
         */
        max_query++;
        egd_query[max_query - 1].fd = accept_fd;
        egd_query[max_query - 1].state = wait_command;
	if (debug) {
          (void)fprintf(stderr, "Opening service fd %d\n",
	  	egd_query[max_query-1].fd);
	}
	/*
	 * Finally, mix information about the connection into the pool.
	 * Maybe its coming from TCP and has some foreign port number
	 * in it. Only count it as '0', because maybe there is no
	 * entropy in it at all.
	 */
	rand_add(&addr, addrlen, 0.0);
      }
    }

  }

}
